from __future__ import annotations

import copy
from typing import cast

from kosong.utils.typing import JsonType

type JsonDict = dict[str, JsonType]


def deref_json_schema(schema: JsonDict) -> JsonDict:
    """Expand local `$ref` entries in a JSON Schema without infinite recursion."""
    # Work on a deep copy so we never mutate the caller's schema.
    full_schema: JsonDict = copy.deepcopy(schema)

    def resolve_pointer(root: JsonDict, pointer: str) -> JsonType:
        """Resolve a JSON Pointer (e.g. ``#/$defs/User``) inside the schema."""
        parts = pointer.lstrip("#/").split("/")
        current: JsonType = root
        try:
            for part in parts:
                if isinstance(current, dict):
                    current = current[part]
                else:
                    raise ValueError
            return current
        except (KeyError, TypeError, ValueError):
            raise ValueError(f"Unable to resolve reference path: {pointer}") from None

    def traverse(node: JsonType, root: JsonDict) -> JsonType:
        """Recursively traverse every node to inline local references."""
        if isinstance(node, dict):
            # Replace local ``$ref`` entries with their referenced payload.
            if "$ref" in node and isinstance(node["$ref"], str):
                ref_path = node["$ref"]
                if ref_path.startswith("#"):
                    # Resolve the local reference target.
                    target = resolve_pointer(root, ref_path)
                    # Recursively inline the target in case it contains more refs.
                    ref = traverse(target, root)
                    if not isinstance(ref, dict):
                        msg = "Local $ref must resolve to a JSON object"
                        raise TypeError(msg)
                    node.pop("$ref")
                    node.update(ref)
                    return node
                else:
                    # Ignore remote references such as http://...
                    return node

            # Traverse the remaining mapping entries.
            return {k: traverse(v, root) for k, v in node.items()}

        elif isinstance(node, list):
            # Traverse list members (e.g. allOf, oneOf, items).
            return [traverse(item, root) for item in node]

        else:
            return node

    # Remove definition buckets to keep the resolved schema minimal.
    resolved = cast(JsonDict, traverse(full_schema, full_schema))

    # Comment these lines if you want to keep the emitted definitions.
    resolved.pop("$defs", None)
    resolved.pop("definitions", None)

    return resolved



import copy
import mimetypes
import os
import uuid
from collections.abc import AsyncIterator, Sequence
from typing import TYPE_CHECKING, Any, Literal, Self, Unpack, cast

import httpx
from openai import AsyncOpenAI, AsyncStream, BaseModel, OpenAIError, omit
from openai._types import RequestFiles, RequestOptions
from openai.types.chat import (
    ChatCompletion,
    ChatCompletionChunk,
    ChatCompletionMessageFunctionToolCall,
    ChatCompletionMessageParam,
    ChatCompletionToolParam,
)
from openai.types.completion_usage import CompletionUsage
from typing_extensions import TypedDict

from kosong.chat_provider import (
    ChatProvider,
    ChatProviderError,
    StreamedMessagePart,
    ThinkingEffort,
    TokenUsage,
)
from kosong.chat_provider.openai_common import convert_error, tool_to_openai
from kosong.message import (
    ContentPart,
    Message,
    TextPart,
    ThinkPart,
    ToolCall,
    ToolCallPart,
    VideoURLPart,
)
from kosong.tooling import Tool

if TYPE_CHECKING:

    def type_check(kimi: "Kimi"):
        _: ChatProvider = kimi


class ThinkingConfig(TypedDict, total=True):
    type: Literal["enabled", "disabled"]


class ExtraBody(TypedDict, total=False, extra_items=Any):
    thinking: ThinkingConfig


class Kimi:
    """
    A chat provider that uses the Kimi API.

    >>> chat_provider = Kimi(model="kimi-k2-turbo-preview", api_key="sk-1234567890")
    >>> chat_provider.name
    'kimi'
    >>> chat_provider.model_name
    'kimi-k2-turbo-preview'
    >>> chat_provider.with_generation_kwargs(temperature=0)._generation_kwargs
    {'temperature': 0}
    >>> chat_provider._generation_kwargs
    {}
    """

    name = "kimi"

    class GenerationKwargs(TypedDict, total=False):
        """
        See https://platform.moonshot.ai/docs/api/chat#request-body.
        """

        max_tokens: int | None
        temperature: float | None
        top_p: float | None
        n: int | None
        presence_penalty: float | None
        frequency_penalty: float | None
        stop: str | list[str] | None
        prompt_cache_key: str | None
        reasoning_effort: str | None
        """Legacy thinking parameter. Use `extra_body.thinking` instead."""
        extra_body: ExtraBody | None

    def __init__(
        self,
        *,
        model: str,
        api_key: str | None = None,
        base_url: str | None = None,
        stream: bool = True,
        **client_kwargs: Any,
    ):
        if api_key is None:
            api_key = os.getenv("KIMI_API_KEY")
        if api_key is None:
            raise ChatProviderError(
                "The api_key client option or the KIMI_API_KEY environment variable is not set"
            )
        if base_url is None:
            base_url = os.getenv("KIMI_BASE_URL", "https://api.moonshot.ai/v1")

        self.model: str = model
        """The name of the model to use."""
        self.stream: bool = stream
        """Whether to generate responses as a stream."""
        self.client: AsyncOpenAI = AsyncOpenAI(
            api_key=api_key,
            base_url=base_url,
            **client_kwargs,
        )
        """The underlying `AsyncOpenAI` client."""
        self._generation_kwargs: Kimi.GenerationKwargs = {}

    @property
    def model_name(self) -> str:
        return self.model

    @property
    def thinking_effort(self) -> ThinkingEffort | None:
        reasoning_effort = self._generation_kwargs.get("reasoning_effort")
        if reasoning_effort is None:
            return None
        match reasoning_effort:
            case "low":
                return "low"
            case "medium":
                return "medium"
            case "high":
                return "high"
            case _:
                return "off"

    async def generate(
        self,
        system_prompt: str,
        tools: Sequence[Tool],
        history: Sequence[Message],
    ) -> "KimiStreamedMessage":
        messages: list[ChatCompletionMessageParam] = []
        if system_prompt:
            messages.append({"role": "system", "content": system_prompt})
        messages.extend(_convert_message(message) for message in history)

        generation_kwargs: dict[str, Any] = {
            # default kimi generation kwargs
            "max_tokens": 32000,
        }
        generation_kwargs.update(self._generation_kwargs)

        try:
            response = await self.client.chat.completions.create(
                model=self.model,
                messages=messages,
                tools=(_convert_tool(tool) for tool in tools),
                stream=self.stream,
                stream_options={"include_usage": True} if self.stream else omit,
                **generation_kwargs,
            )
            return KimiStreamedMessage(response)
        except (OpenAIError, httpx.HTTPError) as e:
            raise convert_error(e) from e

    def with_thinking(self, effort: ThinkingEffort) -> Self:
        match effort:
            case "off":
                reasoning_effort = None
            case "low":
                reasoning_effort = "low"
            case "medium":
                reasoning_effort = "medium"
            case "high":
                reasoning_effort = "high"
        return self.with_generation_kwargs(reasoning_effort=reasoning_effort).with_extra_body(
            {
                "thinking": {
                    "type": "enabled" if effort != "off" else "disabled",
                }
            }
        )

    def with_generation_kwargs(self, **kwargs: Unpack[GenerationKwargs]) -> Self:
        """
        Copy the chat provider, updating the generation kwargs with the given values.

        Returns:
            Self: A new instance of the chat provider with updated generation kwargs.
        """
        new_self = copy.copy(self)
        new_self._generation_kwargs = copy.deepcopy(self._generation_kwargs)
        new_self._generation_kwargs.update(kwargs)
        return new_self

    def with_extra_body(self, extra_body: ExtraBody) -> Self:
        """
        Copy the chat provider, updating the extra_body in generation kwargs.

        Returns:
            Self: A new instance of the chat provider with updated extra_body.
        """
        new_self = copy.copy(self)
        new_self._generation_kwargs = copy.deepcopy(self._generation_kwargs)
        old_extra_body = new_self._generation_kwargs.get("extra_body") or {}
        new_extra_body: ExtraBody = {**old_extra_body, **extra_body}
        new_self._generation_kwargs["extra_body"] = new_extra_body
        return new_self

    @property
    def model_parameters(self) -> dict[str, Any]:
        """
        The parameters of the model to use.

        For tracing/logging purposes.
        """

        model_parameters: dict[str, Any] = {"base_url": str(self.client.base_url)}
        model_parameters.update(self._generation_kwargs)
        return model_parameters

    @property
    def files(self) -> "KimiFiles":
        return KimiFiles(self.client)


class KimiFiles:
    def __init__(self, client: AsyncOpenAI) -> None:
        self._client = client

    async def upload_video(self, *, data: bytes, mime_type: str) -> VideoURLPart:
        """Upload a video to Kimi files API and return a video URL content part."""
        if not mime_type.startswith("video/"):
            raise ChatProviderError(f"Expected a video mime type, got {mime_type}")
        url = await self._upload_file(data=data, mime_type=mime_type, purpose="video")
        return VideoURLPart(video_url=VideoURLPart.VideoURL(url=url))

    async def _upload_file(self, *, data: bytes, mime_type: str, purpose: "KimiFilePurpose") -> str:
        filename = _guess_filename(mime_type)
        files: RequestFiles = {"file": (filename, data, mime_type)}
        options: RequestOptions = {"headers": {"Content-Type": "multipart/form-data"}}
        try:
            response: KimiFileObject = await self._client.post(
                "/files",
                cast_to=KimiFileObject,
                body={"purpose": purpose},
                files=files,
                options=options,
            )
        except (OpenAIError, httpx.HTTPError) as e:
            raise convert_error(e) from e
        return f"ms://{response.id}"


class KimiFileObject(BaseModel):
    id: str


type KimiFilePurpose = Literal["video", "image"]


def _guess_filename(mime_type: str) -> str:
    extension = mimetypes.guess_extension(mime_type) or ".bin"
    return f"upload{extension}"


def _convert_message(message: Message) -> ChatCompletionMessageParam:
    message = message.model_copy(deep=True)
    reasoning_content: str = ""
    content: list[ContentPart] = []
    for part in message.content:
        if isinstance(part, ThinkPart):
            reasoning_content += part.think
        else:
            content.append(part)
    message.content = content
    dumped_message = message.model_dump(exclude_none=True)
    if reasoning_content:
        dumped_message["reasoning_content"] = reasoning_content
    return cast(ChatCompletionMessageParam, dumped_message)


def _convert_tool(tool: Tool) -> ChatCompletionToolParam:
    if tool.name.startswith("$"):
        # Kimi builtin functions start with `$`
        return cast(
            ChatCompletionToolParam,
            {
                "type": "builtin_function",
                "function": {
                    "name": tool.name,
                    # no need to set description and parameters
                },
            },
        )
    else:
        return tool_to_openai(tool)


class KimiStreamedMessage:
    """The streamed message of the Kimi chat provider."""

    def __init__(self, response: ChatCompletion | AsyncStream[ChatCompletionChunk]):
        if isinstance(response, ChatCompletion):
            self._iter = self._convert_non_stream_response(response)
        else:
            self._iter = self._convert_stream_response(response)
        self._id: str | None = None
        self._usage: CompletionUsage | None = None

    def __aiter__(self) -> AsyncIterator[StreamedMessagePart]:
        return self

    async def __anext__(self) -> StreamedMessagePart:
        return await self._iter.__anext__()

    @property
    def id(self) -> str | None:
        return self._id

    @property
    def usage(self) -> TokenUsage | None:
        if self._usage:
            cached = 0
            other_input = self._usage.prompt_tokens
            if hasattr(self._usage, "cached_tokens"):
                # https://platform.moonshot.cn/docs/api/chat#%E8%BF%94%E5%9B%9E%E5%86%85%E5%AE%B9
                # TODO: delete this when Moonshot API becomes compatible with OpenAI API
                cached = getattr(self._usage, "cached_tokens") or 0  # noqa: B009
                other_input -= cached
            elif (
                self._usage.prompt_tokens_details
                and self._usage.prompt_tokens_details.cached_tokens
            ):
                cached = self._usage.prompt_tokens_details.cached_tokens
                other_input -= cached
            return TokenUsage(
                input_other=other_input,
                output=self._usage.completion_tokens,
                input_cache_read=cached,
            )
        return None

    async def _convert_non_stream_response(
        self,
        response: ChatCompletion,
    ) -> AsyncIterator[StreamedMessagePart]:
        self._id = response.id
        self._usage = response.usage
        message = response.choices[0].message
        if reasoning_content := getattr(message, "reasoning_content", None):
            assert isinstance(reasoning_content, str)
            yield ThinkPart(think=reasoning_content)
        if message.content:
            yield TextPart(text=message.content)
        if message.tool_calls:
            for tool_call in message.tool_calls:
                if isinstance(tool_call, ChatCompletionMessageFunctionToolCall):
                    yield ToolCall(
                        id=tool_call.id or str(uuid.uuid4()),
                        function=ToolCall.FunctionBody(
                            name=tool_call.function.name,
                            arguments=tool_call.function.arguments,
                        ),
                    )

    async def _convert_stream_response(
        self,
        response: AsyncIterator[ChatCompletionChunk],
    ) -> AsyncIterator[StreamedMessagePart]:
        try:
            async for chunk in response:
                if chunk.id:
                    self._id = chunk.id
                if usage := extract_usage_from_chunk(chunk):
                    self._usage = usage

                if not chunk.choices:
                    continue

                delta = chunk.choices[0].delta

                # convert thinking content
                if reasoning_content := getattr(delta, "reasoning_content", None):
                    assert isinstance(reasoning_content, str)
                    yield ThinkPart(think=reasoning_content)

                # convert text content
                if delta.content:
                    yield TextPart(text=delta.content)

                # convert tool calls
                for tool_call in delta.tool_calls or []:
                    if not tool_call.function:
                        continue

                    if tool_call.function.name:
                        yield ToolCall(
                            id=tool_call.id or str(uuid.uuid4()),
                            function=ToolCall.FunctionBody(
                                name=tool_call.function.name,
                                arguments=tool_call.function.arguments,
                            ),
                        )
                    elif tool_call.function.arguments:
                        yield ToolCallPart(
                            arguments_part=tool_call.function.arguments,
                        )
                    else:
                        # skip empty tool calls
                        pass
        except (OpenAIError, httpx.HTTPError) as e:
            raise convert_error(e) from e


def extract_usage_from_chunk(chunk: ChatCompletionChunk) -> CompletionUsage | None:
    if chunk.usage:
        return chunk.usage
    if not chunk.choices:
        return None
    choice_dump: dict[str, object] = chunk.choices[0].model_dump()
    raw_usage = choice_dump.get("usage")
    if isinstance(raw_usage, CompletionUsage):
        return raw_usage
    if isinstance(raw_usage, dict):
        return CompletionUsage.model_validate(raw_usage)
    return None


if __name__ == "__main__":

    async def _dev_main():
        chat = Kimi(model="kimi-k2-turbo-preview", stream=False)
        system_prompt = ""
        history = [
            Message(role="user", content="Hello, who is Confucius?"),
        ]
        stream = await chat.with_generation_kwargs(
            temperature=0,
            max_tokens=1000,
        ).generate(system_prompt, [], history)
        async for part in stream:
            print(part.model_dump(exclude_none=True))
        print("id:", stream.id)
        print("usage:", stream.usage)

    import asyncio

    from dotenv import load_dotenv

    load_dotenv()
    asyncio.run(_dev_main())



import asyncio
import inspect
import json
from collections.abc import Iterable
from typing import TYPE_CHECKING, Any, Self

from kosong.message import ToolCall
from kosong.tooling import (
    CallableTool,
    CallableTool2,
    HandleResult,
    Tool,
    ToolResult,
    ToolReturnValue,
    Toolset,
)
from kosong.tooling.error import (
    ToolNotFoundError,
    ToolParseError,
    ToolRuntimeError,
)
from kosong.utils.typing import JsonType

if TYPE_CHECKING:

    def type_check(
        simple: "SimpleToolset",
    ):
        _: Toolset = simple


type ToolType = CallableTool | CallableTool2[Any]
"""The tool type that can be added to the `SimpleToolset`."""


class SimpleToolset:
    """A simple toolset that can handle tool calls concurrently."""

    _tool_dict: dict[str, ToolType]

    def __init__(self, tools: Iterable[ToolType] | None = None):
        """Initialize the simple toolset with an optional iterable of tools."""
        self._tool_dict = {}
        if tools:
            for tool in tools:
                self += tool

    def __iadd__(self, tool: ToolType) -> Self:
        """
        @public
        Add a tool to the toolset.
        """
        return_annotation = inspect.signature(tool.__call__).return_annotation

        # Check if the return annotation is ToolReturnValue
        # Supports both actual type and string annotation (when using
        # `from __future__ import annotations`)
        if return_annotation is ToolReturnValue:
            pass
        elif isinstance(return_annotation, str):
            # String annotation - check if it matches ToolReturnValue
            # Accept any suffix of the full module path, e.g.:
            #   "ToolReturnValue", "tooling.ToolReturnValue", "kosong.tooling.ToolReturnValue"
            full_name = f"{ToolReturnValue.__module__}.ToolReturnValue"
            full_parts = full_name.split(".")
            if not any(
                return_annotation == ".".join(full_parts[i:]) for i in range(len(full_parts))
            ):
                raise TypeError(
                    f"Expected tool `{tool.name}` to return `ToolReturnValue`, "
                    f"but got `{return_annotation}`"
                )
        else:
            raise TypeError(
                f"Expected tool `{tool.name}` to return `ToolReturnValue`, "
                f"but got `{return_annotation}`"
            )

        self._tool_dict[tool.name] = tool
        return self

    def __add__(self, tool: ToolType) -> "SimpleToolset":
        """
        @public
        Return a new toolset with the given tool added.
        """
        new_toolset = SimpleToolset()
        new_toolset._tool_dict = self._tool_dict.copy()
        new_toolset += tool
        return new_toolset

    def add(self, tool: ToolType) -> None:
        """
        @public
        Add a tool to the toolset.
        """
        self += tool

    def remove(self, tool_name: str) -> None:
        """
        @public
        Remove a tool from the toolset.
        """
        if tool_name not in self._tool_dict:
            raise KeyError(f"Tool `{tool_name}` not found in the toolset.")
        del self._tool_dict[tool_name]

    @property
    def tools(self) -> list[Tool]:
        return [tool.base for tool in self._tool_dict.values()]

    def handle(self, tool_call: ToolCall) -> HandleResult:
        if tool_call.function.name not in self._tool_dict:
            return ToolResult(
                tool_call_id=tool_call.id,
                return_value=ToolNotFoundError(tool_call.function.name),
            )

        tool = self._tool_dict[tool_call.function.name]

        try:
            arguments: JsonType = json.loads(tool_call.function.arguments or "{}")
        except json.JSONDecodeError as e:
            return ToolResult(tool_call_id=tool_call.id, return_value=ToolParseError(str(e)))

        async def _call():
            try:
                ret = await tool.call(arguments)
                return ToolResult(tool_call_id=tool_call.id, return_value=ret)
            except Exception as e:
                return ToolResult(tool_call_id=tool_call.id, return_value=ToolRuntimeError(str(e)))

        return asyncio.create_task(_call())



import mcp.types

import kosong.message


def convert_mcp_content(part: mcp.types.ContentBlock) -> kosong.message.ContentPart:
    """Convert MCP content block to kosong message content part.

    Raises:
        ValueError: If the content type or mime type is not supported.
    """
    match part:
        case mcp.types.TextContent(text=text):
            return kosong.message.TextPart(text=text)
        case mcp.types.ImageContent(data=data, mimeType=mimeType):
            return kosong.message.ImageURLPart(
                image_url=kosong.message.ImageURLPart.ImageURL(url=f"data:{mimeType};base64,{data}")
            )

        case mcp.types.AudioContent(data=data, mimeType=mimeType):
            return kosong.message.AudioURLPart(
                audio_url=kosong.message.AudioURLPart.AudioURL(url=f"data:{mimeType};base64,{data}")
            )
        case mcp.types.EmbeddedResource(
            resource=mcp.types.BlobResourceContents(uri=_uri, mimeType=mimeType, blob=blob)
        ):
            mimeType = mimeType or "application/octet-stream"
            if mimeType.startswith("image/"):
                return kosong.message.ImageURLPart(
                    type="image_url",
                    image_url=kosong.message.ImageURLPart.ImageURL(
                        url=f"data:{mimeType};base64,{blob}",
                    ),
                )
            elif mimeType.startswith("audio/"):
                return kosong.message.AudioURLPart(
                    type="audio_url",
                    audio_url=kosong.message.AudioURLPart.AudioURL(
                        url=f"data:{mimeType};base64,{blob}"
                    ),
                )
            elif mimeType.startswith("video/"):
                return kosong.message.VideoURLPart(
                    type="video_url",
                    video_url=kosong.message.VideoURLPart.VideoURL(
                        url=f"data:{mimeType};base64,{blob}"
                    ),
                )

            else:
                raise ValueError(f"Unsupported mime type: {mimeType}")
        case mcp.types.ResourceLink(uri=uri, mimeType=mimeType, description=_description):
            mimeType = mimeType or "application/octet-stream"
            if mimeType.startswith("image/"):
                return kosong.message.ImageURLPart(
                    type="image_url",
                    image_url=kosong.message.ImageURLPart.ImageURL(url=str(uri)),
                )
            elif mimeType.startswith("audio/"):
                return kosong.message.AudioURLPart(
                    type="audio_url",
                    audio_url=kosong.message.AudioURLPart.AudioURL(url=str(uri)),
                )
            elif mimeType.startswith("video/"):
                return kosong.message.VideoURLPart(
                    type="video_url",
                    video_url=kosong.message.VideoURLPart.VideoURL(url=str(uri)),
                )
            else:
                raise ValueError(f"Unsupported mime type: {mimeType}")
        case _:
            raise ValueError(f"Unsupported MCP tool result part: {part}")



from __future__ import annotations

import json
from typing import Any, cast

from kosong.chat_provider import ChatProviderError, StreamedMessagePart, TokenUsage
from kosong.message import (
    AudioURLPart,
    ImageURLPart,
    TextPart,
    ThinkPart,
    ToolCall,
    ToolCallPart,
    VideoURLPart,
)


def parse_echo_script(
    script: str,
) -> tuple[list[StreamedMessagePart], str | None, TokenUsage | None]:
    parts: list[StreamedMessagePart] = []
    message_id: str | None = None
    usage: TokenUsage | None = None

    for lineno, raw_line in enumerate(script.splitlines(), start=1):
        line = raw_line.strip()
        if not line or line.startswith("#") or line.startswith("```"):
            continue
        if line.lower() == "echo":
            continue
        key, sep, payload = line.partition(":")
        if not sep:
            raise ChatProviderError(f"Invalid echo DSL at line {lineno}: {raw_line!r}")

        kind = key.strip().lower()
        payload = payload[1:] if payload.startswith(" ") else payload
        if kind == "id":
            message_id = _strip_quotes(payload.strip())
            continue
        if kind == "usage":
            usage = _parse_usage(payload)
            continue

        part = _parse_part(kind, payload, lineno, raw_line)
        parts.append(part)

    return parts, message_id, usage


def _parse_part(kind: str, payload: str, lineno: int, raw_line: str) -> StreamedMessagePart:
    match kind:
        case "text":
            return TextPart(text=_strip_quotes(payload))
        case "think":
            return ThinkPart(think=_strip_quotes(payload))
        case "image_url":
            url, image_id = _parse_url_payload(payload, kind)
            return ImageURLPart(image_url=ImageURLPart.ImageURL(url=url, id=image_id))
        case "audio_url":
            url, audio_id = _parse_url_payload(payload, kind)
            return AudioURLPart(audio_url=AudioURLPart.AudioURL(url=url, id=audio_id))
        case "video_url":
            url, video_id = _parse_url_payload(payload, kind)
            return VideoURLPart(video_url=VideoURLPart.VideoURL(url=url, id=video_id))
        case "tool_call":
            return _parse_tool_call(payload, lineno, raw_line)
        case "tool_call_part":
            return _parse_tool_call_part(payload)
        case _:
            raise ChatProviderError(
                f"Unknown echo DSL kind '{kind}' at line {lineno}: {raw_line!r}"
            )


def _parse_usage(payload: str) -> TokenUsage:
    mapping = _parse_mapping(payload, context="usage")

    def _int_value(key: str) -> int:
        value = mapping.get(key, 0)
        try:
            return int(value)
        except (TypeError, ValueError):
            raise ChatProviderError(
                f"Usage field '{key}' must be an integer, got {value!r}"
            ) from None

    return TokenUsage(
        input_other=_int_value("input_other"),
        output=_int_value("output"),
        input_cache_read=_int_value("input_cache_read"),
        input_cache_creation=_int_value("input_cache_creation"),
    )


def _parse_url_payload(payload: str, kind: str) -> tuple[str, str | None]:
    value = _parse_value(payload)
    if isinstance(value, dict):
        mapping = cast(dict[str, Any], value)
        url = mapping.get("url")
        if not isinstance(url, str):
            raise ChatProviderError(f"{kind} requires a url field, got {mapping!r}")
        content_id = mapping.get("id")
        if content_id is not None and not isinstance(content_id, str):
            raise ChatProviderError(f"{kind} id must be a string when provided.")
        return url, content_id
    if not isinstance(value, str):
        raise ChatProviderError(f"{kind} expects url string or object, got {value!r}")
    return value, None


def _parse_tool_call(payload: str, lineno: int, raw_line: str) -> ToolCall:
    mapping = _parse_mapping(payload, context="tool_call")
    function = mapping.get("function") if isinstance(mapping.get("function"), dict) else None

    tool_call_id = mapping.get("id")
    name = mapping.get("name") or (function.get("name") if function else None)
    arguments = mapping.get("arguments")
    extras = mapping.get("extras")

    if function:
        if arguments is None:
            arguments = function.get("arguments")
        if extras is None:
            extras = function.get("extras")

    if not isinstance(tool_call_id, str) or not isinstance(name, str):
        raise ChatProviderError(
            f"tool_call requires string id and name at line {lineno}: {raw_line!r}"
        )

    if arguments is not None and not isinstance(arguments, str):
        raise ChatProviderError(
            f"tool_call.arguments must be a string at line {lineno}, got {type(arguments).__name__}"
        )

    return ToolCall(
        id=tool_call_id,
        function=ToolCall.FunctionBody(name=name, arguments=arguments),
        extras=cast(dict[str, Any], extras) if isinstance(extras, dict) else None,
    )


def _parse_tool_call_part(payload: str) -> ToolCallPart:
    value = _parse_value(payload)
    if isinstance(value, dict):
        value = cast(dict[str, Any], value)
        arguments_part: Any | None = value.get("arguments_part")
    else:
        arguments_part = value
    if isinstance(arguments_part, (dict, list)):
        arguments_part = json.dumps(arguments_part, separators=(",", ":"))
    return ToolCallPart(arguments_part=None if arguments_part in (None, "") else arguments_part)


def _parse_mapping(raw: str, *, context: str) -> dict[str, Any]:
    raw = raw.strip()
    try:
        loaded = json.loads(raw)
    except json.JSONDecodeError:
        loaded = None
    if isinstance(loaded, dict):
        return cast(dict[str, Any], loaded)
    if loaded is not None:
        raise ChatProviderError(f"{context} payload must be an object, got {loaded!r}")

    mapping: dict[str, Any] = {}
    for token in raw.replace(",", " ").split():
        if not token:
            continue
        if "=" not in token:
            raise ChatProviderError(f"Invalid token '{token}' in {context} payload.")
        key, value = token.split("=", 1)
        mapping[key.strip()] = _parse_value(value.strip())

    if not mapping:
        raise ChatProviderError(f"{context} payload cannot be empty.")
    return mapping


def _parse_value(raw: str) -> Any:
    raw = raw.strip()
    if not raw:
        return None
    lowered = raw.lower()
    if lowered in {"null", "none"}:
        return None
    try:
        return json.loads(raw)
    except json.JSONDecodeError:
        return _strip_quotes(raw)


def _strip_quotes(value: str) -> str:
    if len(value) >= 2 and value[0] == value[-1] and value[0] in {"'", '"'}:
        return value[1:-1]
    return value



from __future__ import annotations

import copy
from collections.abc import AsyncIterator, Sequence
from typing import TYPE_CHECKING, Self

from kosong.chat_provider import (
    ChatProvider,
    ChatProviderError,
    StreamedMessage,
    StreamedMessagePart,
    ThinkingEffort,
    TokenUsage,
)
from kosong.chat_provider.echo.dsl import parse_echo_script
from kosong.message import Message
from kosong.tooling import Tool

if TYPE_CHECKING:

    def type_check(echo: EchoChatProvider):
        _: ChatProvider = echo


class EchoChatProvider:
    """
    A test-only chat provider that streams parts described by a tiny DSL.

    The DSL lives in the content of the last message in `history` and is made of lines in the
    form `kind: payload`. Empty lines, comment lines starting with `#`, and markdown fences
    starting with ``` are ignored. Supported kinds:

    - `id`: sets the streamed message id.
    - `usage`: token usage, e.g. `usage: {"input_other": 10, "output": 2}` or
      `usage: input_other=1 output=2 input_cache_read=3`.
    - `text`: a text chunk.
    - `think`: a thinking chunk.
    - `image_url`: either a raw URL or `{"url": "...", "id": "opt"}`.
    - `audio_url`: either a raw URL or `{"url": "...", "id": "opt"}`.
    - `video_url`: either a raw URL or `{"url": "...", "id": "opt"}`.
    - `tool_call`: a JSON or key/value object. Fields: `id`, `name` (or `function.name`),
      optional `arguments`/`function.arguments`, optional `extras`.
    - `tool_call_part`: a string/JSON with `arguments_part`; `null` becomes `None`.

    Example:

    ```
    id: echo-42
    usage: {"input_other": 10, "output": 2}
    think: thinking...
    text: Hello,
    text:  world!
    image_url: {"url": "https://example.com/image.png", "id": "img-1"}
    tool_call: {"id": "call-1", "name": "search", "arguments": "{\\"query"}
    tool_call_part: {"arguments_part": "\\": \\"what time is"}
    tool_call_part: {"arguments_part": " it?\\"}"}
    ```
    """

    name = "echo"

    @property
    def model_name(self) -> str:
        return "echo"

    @property
    def thinking_effort(self) -> ThinkingEffort | None:
        return None

    async def generate(
        self,
        system_prompt: str,
        tools: Sequence[Tool],
        history: Sequence[Message],
    ) -> EchoStreamedMessage:
        if not history:
            raise ChatProviderError("EchoChatProvider requires at least one message in history.")
        if history[-1].role != "user":
            raise ChatProviderError("EchoChatProvider expects the last history message to be user.")

        script_text = history[-1].extract_text()
        parts, message_id, usage = parse_echo_script(script_text)
        if not parts:
            raise ChatProviderError("EchoChatProvider DSL produced no streamable parts.")
        return EchoStreamedMessage(parts=parts, message_id=message_id, usage=usage)

    def with_thinking(self, effort: ThinkingEffort) -> Self:
        # Thinking effort is irrelevant to the echo provider; return a shallow copy to
        # satisfy the protocol and keep the instance immutable.
        return copy.copy(self)


class EchoStreamedMessage(StreamedMessage):
    """Streamed message for EchoChatProvider."""

    def __init__(
        self,
        *,
        parts: list[StreamedMessagePart],
        message_id: str | None,
        usage: TokenUsage | None,
    ):
        self._iter = self._to_stream(parts)
        self._id = message_id
        self._usage = usage

    def __aiter__(self) -> AsyncIterator[StreamedMessagePart]:
        return self

    async def __anext__(self) -> StreamedMessagePart:
        return await self._iter.__anext__()

    async def _to_stream(
        self, parts: list[StreamedMessagePart]
    ) -> AsyncIterator[StreamedMessagePart]:
        for part in parts:
            yield part

    @property
    def id(self) -> str | None:
        return self._id

    @property
    def usage(self) -> TokenUsage | None:
        return self._usage



try:
    from google import genai as _  # noqa: F401
except ModuleNotFoundError as exc:
    raise ModuleNotFoundError(
        "Google Gemini support requires the optional dependency 'google-genai'. "
        'Install with `pip install "kosong[contrib]"`.'
    ) from exc

import base64
import copy
import json
import mimetypes
from collections.abc import AsyncIterator, Sequence
from typing import TYPE_CHECKING, Any, Self, TypedDict, Unpack, cast

import httpx
from google import genai
from google.genai import client as genai_client
from google.genai import errors as genai_errors
from google.genai.types import (
    Content,
    FunctionCall,
    FunctionDeclaration,
    FunctionResponse,
    FunctionResponsePart,
    GenerateContentConfig,
    GenerateContentResponse,
    GenerateContentResponseUsageMetadata,
    HttpOptions,
    Part,
    ThinkingConfig,
    ThinkingLevel,
    Tool,
    ToolConfig,
)

from kosong.chat_provider import (
    APIStatusError,
    APITimeoutError,
    ChatProvider,
    ChatProviderError,
    StreamedMessagePart,
    ThinkingEffort,
    TokenUsage,
)
from kosong.message import (
    AudioURLPart,
    ContentPart,
    ImageURLPart,
    Message,
    TextPart,
    ThinkPart,
    ToolCall,
)
from kosong.tooling import Tool as KosongTool
from kosong.tooling import ToolReturnValue

if TYPE_CHECKING:

    def type_check(google_genai: "GoogleGenAI"):
        _: ChatProvider = google_genai


class GoogleGenAI:
    """
    Chat provider backed by Google's Gemini API.
    """

    name = "google_genai"

    class GenerationKwargs(TypedDict, total=False):
        max_output_tokens: int | None
        temperature: float | None
        top_k: int | None
        top_p: float | None
        # Thinking configuration for supported models
        thinking_config: ThinkingConfig | None
        # Tool configuration
        tool_config: ToolConfig | None
        # Extra headers
        http_options: HttpOptions | None

    def __init__(
        self,
        *,
        model: str,
        api_key: str | None = None,
        base_url: str | None = None,
        stream: bool = True,
        vertexai: bool | None = None,
        **client_kwargs: Any,
    ):
        self._model = model
        self._stream = stream
        self._base_url = base_url
        self._client: genai_client.Client = genai.Client(
            http_options=HttpOptions(base_url=base_url),
            api_key=api_key,
            vertexai=vertexai,
            **client_kwargs,
        )
        self._generation_kwargs: GoogleGenAI.GenerationKwargs = {}

    @property
    def model_name(self) -> str:
        return self._model

    @property
    def thinking_effort(self) -> "ThinkingEffort | None":
        thinking_config = self._generation_kwargs.get("thinking_config")
        if thinking_config is None:
            return None

        # For gemini-3 models that use thinking_level
        thinking_level = thinking_config.thinking_level
        if thinking_level is not None:
            match thinking_level:
                case ThinkingLevel.LOW | ThinkingLevel.MINIMAL:
                    return "low"
                case ThinkingLevel.MEDIUM:
                    return "medium"
                case ThinkingLevel.HIGH:
                    return "high"
                case _:
                    return None

        # For other models that use thinking_budget
        thinking_budget = thinking_config.thinking_budget
        if thinking_budget is not None:
            if thinking_budget == 0:
                return "off"
            if thinking_budget <= 1024:
                return "low"
            if thinking_budget <= 4096:
                return "medium"
            return "high"
        return None

    async def generate(
        self,
        system_prompt: str,
        tools: Sequence[KosongTool],
        history: Sequence[Message],
    ) -> "GoogleGenAIStreamedMessage":
        contents = messages_to_google_genai_contents(history)

        config = GenerateContentConfig(**self._generation_kwargs)
        config.system_instruction = system_prompt
        config.tools = [tool_to_google_genai(tool) for tool in tools]

        try:
            if self._stream:
                stream_response = await self._client.aio.models.generate_content_stream(  # type: ignore[reportUnknownMemberType]
                    model=self._model,
                    contents=contents,  # type: ignore[reportArgumentType]
                    config=config,
                )
                return GoogleGenAIStreamedMessage(stream_response)
            else:
                response = await self._client.aio.models.generate_content(  # type: ignore[reportUnknownMemberType]
                    model=self._model,
                    contents=contents,  # type: ignore[reportArgumentType]
                    config=config,
                )
                return GoogleGenAIStreamedMessage(response)
        except Exception as e:  # genai_errors.APIError and others
            raise _convert_error(e) from e

    def with_thinking(self, effort: "ThinkingEffort") -> Self:
        thinking_config = ThinkingConfig(include_thoughts=True)

        # Map thinking effort to budget tokens
        if "gemini-3" in self._model:
            match effort:
                case "off":
                    # use default thinking config
                    pass
                case "low":
                    thinking_config.thinking_level = ThinkingLevel.LOW
                case "medium":
                    # FIXME: medium not supported yet, use high
                    thinking_config.thinking_level = ThinkingLevel.HIGH
                case "high":
                    thinking_config.thinking_level = ThinkingLevel.HIGH
        else:
            match effort:
                case "off":
                    thinking_config.thinking_budget = 0
                    thinking_config.include_thoughts = False
                case "low":
                    thinking_config.thinking_budget = 1024
                    thinking_config.include_thoughts = True
                case "medium":
                    thinking_config.thinking_budget = 4096
                    thinking_config.include_thoughts = True
                case "high":
                    thinking_config.thinking_budget = 32_000
                    thinking_config.include_thoughts = True

        return self.with_generation_kwargs(thinking_config=thinking_config)

    def with_generation_kwargs(self, **kwargs: Unpack[GenerationKwargs]) -> Self:
        """
        Copy the chat provider, updating the generation kwargs with the given values.

        Returns:
            Self: A new instance of the chat provider with updated generation kwargs.
        """
        new_self = copy.copy(self)
        new_self._generation_kwargs = copy.deepcopy(self._generation_kwargs)
        new_self._generation_kwargs.update(kwargs)
        return new_self

    @property
    def model_parameters(self) -> dict[str, Any]:
        """
        The parameters of the model to use.

        For tracing/logging purposes.
        """
        return {
            "model": self._model,
            "base_url": self._base_url,
            **self._generation_kwargs,
        }


class GoogleGenAIStreamedMessage:
    def __init__(self, response: GenerateContentResponse | AsyncIterator[GenerateContentResponse]):
        if isinstance(response, GenerateContentResponse):
            self._iter = self._convert_non_stream_response(response)
        else:
            self._iter = self._convert_stream_response(response)
        self._id: str | None = None
        self._usage: GenerateContentResponseUsageMetadata | None = None

    def __aiter__(self) -> AsyncIterator[StreamedMessagePart]:
        return self

    async def __anext__(self) -> StreamedMessagePart:
        return await self._iter.__anext__()

    @property
    def id(self) -> str | None:
        return self._id

    @property
    def usage(self) -> TokenUsage | None:
        if self._usage is None:
            return None
        return TokenUsage(
            input_other=self._usage.prompt_token_count or 0,
            output=self._usage.candidates_token_count or 0,
            input_cache_read=self._usage.cached_content_token_count or 0,
            input_cache_creation=0,
        )

    async def _convert_non_stream_response(
        self,
        response: GenerateContentResponse,
    ) -> AsyncIterator[StreamedMessagePart]:
        # Extract usage information
        if response.usage_metadata:
            self._usage = response.usage_metadata
        # Extract ID if available
        if response.response_id is not None:
            self._id = response.response_id

        # Process candidates
        for candidate in response.candidates or []:
            parts = candidate.content.parts if candidate.content else None
            if not parts:
                continue
            for part in parts:
                async for message_part in self._process_part_async(part):
                    yield message_part

    async def _convert_stream_response(
        self,
        response_stream: AsyncIterator[GenerateContentResponse],
    ) -> AsyncIterator[StreamedMessagePart]:
        try:
            async for response in response_stream:
                # Extract ID from first response
                if not self._id and response.response_id is not None:
                    self._id = response.response_id

                # Extract usage information
                if response.usage_metadata:
                    self._usage = response.usage_metadata

                # Process candidates
                for candidate in response.candidates or []:
                    parts = candidate.content.parts if candidate.content else None
                    if not parts:
                        continue
                    for part in parts:
                        async for message_part in self._process_part_async(part):
                            yield message_part
        except genai_errors.APIError as exc:
            raise _convert_error(exc) from exc

    def _process_part(self, part: Part):
        """Process a single part and yield message components (synchronous generator).

        Handles different part types from Gemini API:
        - synthetic thinking parts (part.thought is True)
        - encrypted thinking parts (part.thought_signature is not None)
        - text parts
        - function calls
        """
        if part.thought:
            # Synthetic thinking part
            if part.text:
                yield ThinkPart(think=part.text)
        elif part.text:
            # Regular text part
            yield TextPart(text=part.text)
        elif part.function_call:
            func_call = part.function_call
            if func_call.name is None:
                # Skip function calls without a name
                return
            id_ = func_call.id if func_call.id is not None else f"{id(func_call)}"
            tool_call_id = f"{func_call.name}_{id_}"
            # Gemini uses thought_signature to store the encrypted thinking signature.
            # part.thought is synthetic
            # See: https://colab.research.google.com/github/GoogleCloudPlatform/generative-ai/blob/main/gemini/thinking/intro_thought_signatures.ipynb
            thought_signature_b64 = (
                base64.b64encode(part.thought_signature).decode("ascii")
                if part.thought_signature
                else None
            )
            yield ToolCall(
                id=tool_call_id,
                function=ToolCall.FunctionBody(
                    name=func_call.name,
                    arguments=json.dumps(func_call.args) if func_call.args else "{}",
                ),
                extras={
                    "thought_signature_b64": thought_signature_b64,
                }
                if thought_signature_b64
                else None,
            )

    async def _process_part_async(self, part: Part) -> AsyncIterator[StreamedMessagePart]:
        """Async wrapper for _process_part."""
        for message_part in self._process_part(part):
            yield message_part


def tool_to_google_genai(tool: KosongTool) -> Tool:
    """Convert a Kosong tool to GoogleGenAI tool format."""
    # Kosong already validates parameters as JSON Schema format via jsonschema
    # The google-genai SDK accepts dict format and internally converts to Schema
    return Tool(
        function_declarations=[
            FunctionDeclaration(
                name=tool.name,
                description=tool.description,
                parameters=tool.parameters,  # type: ignore[arg-type] # GoogleGenAI accepts dict
            )
        ]
    )


def _image_url_part_to_google_genai(part: ImageURLPart) -> Part:
    """Convert an image URL part to GoogleGenAI format."""
    url = part.image_url.url

    # Handle data URLs
    if url.startswith("data:"):
        # data:[<media-type>][;base64],<data>
        res = url[5:].split(";base64,", 1)
        if len(res) != 2:
            raise ChatProviderError(f"Invalid data URL for image: {url}")

        media_type, data_b64 = res
        if media_type not in ("image/png", "image/jpeg", "image/gif", "image/webp"):
            raise ChatProviderError(
                f"Unsupported media type for base64 image: {media_type}, url: {url}"
            )

        # Decode base64 string to bytes
        data_bytes = base64.b64decode(data_b64)
        return Part.from_bytes(data=data_bytes, mime_type=media_type)
    else:
        # For regular URLs, try to download the image and convert to bytes
        mime_type, _ = mimetypes.guess_type(url)
        if not mime_type or not mime_type.startswith("image/"):
            # Default to image/png if we can't detect or it's not an image type
            mime_type = "image/png"
        response = httpx.get(url).raise_for_status()
        data_bytes = response.content
        return Part.from_bytes(data=data_bytes, mime_type=mime_type)


def _audio_url_part_to_google_genai(part: AudioURLPart) -> Part:
    """Convert an audio URL part to GoogleGenAI format."""
    url = part.audio_url.url

    # Handle data URLs
    if url.startswith("data:"):
        # data:[<media-type>][;base64],<data>
        res = url[5:].split(";base64,", 1)
        if len(res) != 2:
            raise ChatProviderError(f"Invalid data URL for audio: {url}")

        media_type, data_b64 = res
        # Supported audio formats for GoogleGenAI
        supported_audio_types = (
            "audio/wav",
            "audio/mp3",
            "audio/aiff",
            "audio/aac",
            "audio/ogg",
            "audio/flac",
        )
        if media_type not in supported_audio_types:
            error_msg = (
                f"Unsupported media type for base64 audio: {media_type}, url: {url}. "
                f"Supported types: {supported_audio_types}"
            )
            raise ChatProviderError(error_msg)

        # Decode base64 string to bytes
        data_bytes = base64.b64decode(data_b64)
        return Part.from_bytes(data=data_bytes, mime_type=media_type)
    else:
        # Fetch the audio and convert to bytes
        mime_type, _ = mimetypes.guess_type(url)
        if not mime_type or not mime_type.startswith("audio/"):
            # Default to audio/mp3 if we can't detect or it's not an audio type
            mime_type = "audio/mp3"
        response = httpx.get(url).raise_for_status()
        data_bytes = response.content
        return Part.from_bytes(data=data_bytes, mime_type=mime_type)


def _tool_result_to_response_and_parts(
    parts: list[ContentPart],
) -> tuple[dict[str, str], list[FunctionResponsePart]]:
    """Convert tool response content to Gemini function response format."""
    genai_parts: list[FunctionResponsePart] = []
    response: str = ""

    for part in parts:
        if isinstance(part, TextPart):
            if part.text:
                response += part.text
        elif isinstance(part, ImageURLPart):
            genai_parts.append(FunctionResponsePart.from_uri(file_uri=part.image_url.url))
        elif isinstance(part, AudioURLPart):
            genai_parts.append(FunctionResponsePart.from_uri(file_uri=part.audio_url.url))
        else:
            # Skip unsupported parts (like ThinkPart, etc.)
            continue

    return {"output": response}, genai_parts


def _tool_call_id_to_name(tool_call_id: str, tool_name_by_id: dict[str, str]) -> str:
    """Resolve Gemini `FunctionResponse.name` from a tool_call_id."""
    if tool_call_id in tool_name_by_id:
        return tool_name_by_id[tool_call_id]
    # Fallback for older ids of the form "{tool_name}_{id}".
    return tool_call_id.split("_", 1)[0]


def _tool_message_to_function_response_part(
    message: Message,
    *,
    tool_name_by_id: dict[str, str],
) -> Part:
    if message.role != "tool":  # pragma: no cover - defensive guard
        raise ChatProviderError("Expected a tool message.")
    if message.tool_call_id is None:
        raise ChatProviderError("Tool response is missing `tool_call_id`.")

    response_data, tool_result_parts = _tool_result_to_response_and_parts(message.content)
    return Part(
        function_response=FunctionResponse(
            id=message.tool_call_id,
            name=_tool_call_id_to_name(message.tool_call_id, tool_name_by_id),
            response=response_data,
            parts=tool_result_parts,
        )
    )


def _tool_messages_to_google_genai_content(
    messages: Sequence[Message],
    *,
    tool_name_by_id: dict[str, str],
    expected_tool_call_ids: Sequence[str] | None = None,
    require_all_expected: bool = False,
) -> Content:
    """Pack one-or-more tool results into a single Gemini "user" turn.

    VertexAI-backed Gemini enforces that, for a tool-calling turn, the next
    turn contains the same number of `functionResponse` parts as the preceding
    `functionCall` parts. Packing multiple tool results into a single "user"
    Content keeps us compliant and avoids ordering issues from parallel tool
    execution.
    """
    if not messages:
        raise ChatProviderError("Expected at least one tool message.")

    expected_index: dict[str, int] = (
        {tool_call_id: i for i, tool_call_id in enumerate(expected_tool_call_ids)}
        if expected_tool_call_ids is not None
        else {}
    )
    seen_tool_call_ids: set[str] = set()
    indexed_messages = list(enumerate(messages))
    indexed_messages.sort(
        key=lambda t: (expected_index.get(cast(str, t[1].tool_call_id), 10**9), t[0])
    )

    parts: list[Part] = []
    actual_tool_call_ids: list[str] = []
    for _, message in indexed_messages:
        if message.tool_call_id is None:
            raise ChatProviderError("Tool response is missing `tool_call_id`.")
        if message.tool_call_id in seen_tool_call_ids:
            raise ChatProviderError(f"Duplicate tool response for id: {message.tool_call_id}")
        seen_tool_call_ids.add(message.tool_call_id)
        actual_tool_call_ids.append(message.tool_call_id)
        parts.append(
            _tool_message_to_function_response_part(message, tool_name_by_id=tool_name_by_id)
        )

    if expected_tool_call_ids is not None and require_all_expected:
        expected_set = set(expected_tool_call_ids)
        missing = [
            tool_call_id
            for tool_call_id in expected_tool_call_ids
            if tool_call_id not in seen_tool_call_ids
        ]
        extra = [
            tool_call_id
            for tool_call_id in actual_tool_call_ids
            if tool_call_id not in expected_set
        ]
        if missing:
            raise ChatProviderError(f"Missing tool responses for ids: {missing}")
        if extra:
            raise ChatProviderError(f"Unexpected tool responses for ids: {extra}")

    return Content(role="user", parts=parts)


def messages_to_google_genai_contents(messages: Sequence[Message]) -> list[Content]:
    """Convert internal messages into a Gemini contents list.

    Tool results for a tool-calling turn are packed into a single "user" message
    with N `functionResponse` parts matching the preceding "model" message's
    N `functionCall` parts. This avoids ordering issues from parallel tool
    execution and satisfies VertexAI's stricter validation.
    """
    contents: list[Content] = []
    tool_name_by_id: dict[str, str] = {}

    i = 0
    while i < len(messages):
        message = messages[i]

        if message.role == "assistant" and message.tool_calls:
            contents.append(message_to_google_genai(message))
            expected_tool_call_ids: list[str] = []
            for tool_call in message.tool_calls:
                tool_name_by_id[tool_call.id] = tool_call.function.name
                expected_tool_call_ids.append(tool_call.id)

            # Collect consecutive tool messages that correspond to this turn.
            j = i + 1
            tool_messages: list[Message] = []
            while j < len(messages) and messages[j].role == "tool":
                tool_messages.append(messages[j])
                j += 1

            if tool_messages:
                contents.append(
                    _tool_messages_to_google_genai_content(
                        tool_messages,
                        tool_name_by_id=tool_name_by_id,
                        expected_tool_call_ids=expected_tool_call_ids,
                        require_all_expected=True,
                    )
                )
                i = j
                continue

            i += 1
            continue

        if message.role == "tool":
            # Tool message without an immediately preceding tool-calling assistant
            # message (e.g. truncated history). Convert it best-effort.
            contents.append(
                _tool_messages_to_google_genai_content([message], tool_name_by_id=tool_name_by_id)
            )
            i += 1
            continue

        contents.append(message_to_google_genai(message))
        if message.role == "assistant" and message.tool_calls:
            for tool_call in message.tool_calls:
                tool_name_by_id[tool_call.id] = tool_call.function.name
        i += 1

    return contents


def message_to_google_genai(message: Message) -> Content:
    """Convert a single internal message into GoogleGenAI wire format."""
    role = message.role

    if role == "tool":
        raise ChatProviderError(
            "Tool messages must be converted via messages_to_google_genai_contents "
            "to preserve tool-call ordering and tool-response packing."
        )

    # GoogleGenAI uses: "user" and "model" (not "assistant")
    google_genai_role = "model" if role == "assistant" else role
    parts: list[Part] = []

    # Handle content parts
    for part in message.content:
        if isinstance(part, TextPart):
            parts.append(Part.from_text(text=part.text))
        elif isinstance(part, ImageURLPart):
            parts.append(_image_url_part_to_google_genai(part))
        elif isinstance(part, AudioURLPart):
            parts.append(_audio_url_part_to_google_genai(part))
        elif isinstance(part, ThinkPart):
            # Note: skip part.thought because it is synthetic
            continue
        else:
            # Skip unsupported parts
            continue

    # Handle tool calls for assistant messages
    for tool_call in message.tool_calls or []:
        if tool_call.function.arguments:
            try:
                parsed_arguments = json.loads(tool_call.function.arguments)
            except json.JSONDecodeError as exc:  # pragma: no cover - defensive guard
                raise ChatProviderError("Tool call arguments must be valid JSON.") from exc
            if not isinstance(parsed_arguments, dict):
                raise ChatProviderError("Tool call arguments must be a JSON object.")
            args = cast(dict[str, object], parsed_arguments)
        else:
            args = {}

        function_call = FunctionCall(
            id=tool_call.id,
            name=tool_call.function.name,
            args=args,
        )
        function_call_part = Part(function_call=function_call)
        # Add thought_signature back to function_call
        if tool_call.extras and "thought_signature_b64" in tool_call.extras:
            function_call_part.thought_signature = base64.b64decode(
                cast(str, tool_call.extras["thought_signature_b64"])
            )
        parts.append(function_call_part)

    return Content(role=google_genai_role, parts=parts)


def _convert_error(error: Exception) -> ChatProviderError:
    """Convert a GoogleGenAI error to a Kosong chat provider error."""
    # Handle specific GoogleGenAI error types with detailed status code mapping
    if isinstance(error, genai_errors.ClientError):
        # 4xx client errors
        status_code = getattr(error, "code", 400)
        if status_code == 401:
            return APIStatusError(401, f"Authentication failed: {error}")
        elif status_code == 403:
            return APIStatusError(403, f"Permission denied: {error}")
        elif status_code == 429:
            return APIStatusError(429, f"Rate limit exceeded: {error}")
        return APIStatusError(status_code, str(error))
    elif isinstance(error, genai_errors.ServerError):
        # 5xx server errors
        status_code = getattr(error, "code", 500)
        return APIStatusError(status_code, f"Server error: {error}")
    elif isinstance(error, genai_errors.APIError):
        # Generic API errors
        status_code = getattr(error, "code", 500)
        return APIStatusError(status_code, str(error))
    elif isinstance(error, TimeoutError):
        return APITimeoutError(f"Request timed out: {error}")
    else:
        # Fallback for unexpected errors
        return ChatProviderError(f"Unexpected GoogleGenAI error: {error}")


if __name__ == "__main__":

    async def main():
        import os
        from typing import override

        from pydantic import BaseModel

        import kosong
        from kosong.tooling import CallableTool2, ToolOk
        from kosong.tooling.simple import SimpleToolset

        chat = GoogleGenAI(
            model="gemini-3-pro-preview",
            vertexai=True,
            api_key=os.getenv("VERTEXAI_API_KEY"),
        ).with_thinking("high")
        system_prompt = "You are a helpful assistant."

        class GetWeatherParams(BaseModel):
            city: str

        class GetWeather(CallableTool2[GetWeatherParams]):
            name: str = "get_weather"
            description: str = "Get the weather of a city"
            params: type[GetWeatherParams] = GetWeatherParams

            @override
            async def __call__(self, params: GetWeatherParams) -> ToolReturnValue:
                return ToolOk(output="Sunny")

        toolset = SimpleToolset()
        toolset += GetWeather()
        history = [
            Message(
                role="user",
                content=(
                    "What's the weather like in Beijing and Shanghai? "
                    "Spawn parallel tool calls to get the answer."
                ),
            )
        ]
        result = await kosong.step(chat, system_prompt, toolset, history)
        tool_results = await result.tool_results()

        assistant_message = result.message
        tool_messages = [
            Message(role="tool", content=tr.return_value.output, tool_call_id=tr.tool_call_id)
            for tr in tool_results
        ]
        history.extend([assistant_message] + tool_messages)

        async for part in await chat.generate(system_prompt, toolset.tools, history):
            print(part.model_dump(exclude_none=True))

    import asyncio

    from dotenv import load_dotenv

    load_dotenv()
    asyncio.run(main())



from __future__ import annotations

import copy
import json
from collections import deque
from collections.abc import AsyncIterator, Iterable, Sequence
from typing import TYPE_CHECKING, Self

from kosong.chat_provider import (
    ChatProvider,
    ChatProviderError,
    StreamedMessage,
    StreamedMessagePart,
    ThinkingEffort,
    TokenUsage,
)
from kosong.chat_provider.echo.dsl import parse_echo_script
from kosong.message import Message
from kosong.tooling import Tool

if TYPE_CHECKING:

    def type_check(scripted: ScriptedEchoChatProvider):
        _: ChatProvider = scripted


class ScriptedEchoChatProvider:
    """
    A test-only chat provider that consumes a queue of echo DSL scripts per call.
    """

    name = "scripted_echo"

    def __init__(self, scripts: Iterable[str], *, trace: bool = False):
        self._scripts = deque(scripts)
        self._turn = 0
        self._trace = trace

    @property
    def model_name(self) -> str:
        return "scripted_echo"

    @property
    def thinking_effort(self) -> ThinkingEffort | None:
        return None

    async def generate(
        self,
        system_prompt: str,
        tools: Sequence[Tool],
        history: Sequence[Message],
    ) -> ScriptedEchoStreamedMessage:
        if not self._scripts:
            raise ChatProviderError(f"ScriptedEchoChatProvider exhausted at turn {self._turn + 1}.")
        script_text = self._scripts.popleft()
        if self._trace:
            script_json = json.dumps(script_text)
            print(f"SCRIPTED_ECHO TURN {self._turn + 1}: {script_json}")
        self._turn += 1
        parts, message_id, usage = parse_echo_script(script_text)
        if not parts:
            raise ChatProviderError("ScriptedEchoChatProvider DSL produced no streamable parts.")
        return ScriptedEchoStreamedMessage(parts=parts, message_id=message_id, usage=usage)

    def with_thinking(self, effort: ThinkingEffort) -> Self:
        copied = copy.copy(self)
        copied._scripts = deque(self._scripts)
        return copied


class ScriptedEchoStreamedMessage(StreamedMessage):
    """Streamed message for ScriptedEchoChatProvider."""

    def __init__(
        self,
        *,
        parts: list[StreamedMessagePart],
        message_id: str | None,
        usage: TokenUsage | None,
    ):
        self._iter = self._to_stream(parts)
        self._id = message_id
        self._usage = usage

    def __aiter__(self) -> AsyncIterator[StreamedMessagePart]:
        return self

    async def __anext__(self) -> StreamedMessagePart:
        return await self._iter.__anext__()

    async def _to_stream(
        self, parts: list[StreamedMessagePart]
    ) -> AsyncIterator[StreamedMessagePart]:
        for part in parts:
            yield part

    @property
    def id(self) -> str | None:
        return self._id

    @property
    def usage(self) -> TokenUsage | None:
        return self._usage



from .echo import EchoChatProvider, EchoStreamedMessage
from .scripted_echo import ScriptedEchoChatProvider, ScriptedEchoStreamedMessage

__all__ = [
    "EchoChatProvider",
    "EchoStreamedMessage",
    "ScriptedEchoChatProvider",
    "ScriptedEchoStreamedMessage",
]
